/**
 * C library for acessing TCB services.
 * 
 * author: Paulo Sousa (pjsousa@di.fc.ul.pt)
 * date: 21/03/2003
 */

#include "stdafx.h"
#include "tcb_services_library.h"
#include "private_functions_and_structures.h"
#include "Winsock.h"

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}


// This is an example of an exported variable
TCB_SERVICES_LIBRARY_API int nTcb_services_library=0;

// This is an example of an exported function.
TCB_SERVICES_LIBRARY_API int fnTcb_services_library(void)
{
	return 42;
}


//used for communication TCB DLL<->TCB through memory mapping
service_request_response_t *srrt=NULL;
//used for communication TCB DLL wait_info()<-TCB DistFD service 
//through memory mapping
fail_info_to_app_t *fi=NULL;
//event used by TCB to wait for our requests
HANDLE tcb_wait_event;
//event used for waiting for TCB answers
HANDLE dll_wait_event;
//event used for waiting for TCB distributed 
//timing failure detection service info
HANDLE wait_info_event;
//handles for memory mapped
HANDLE fmap,fmap1;
//indicates if there is a session opened
bool opened=FALSE;
//allows to detect if a session was opened by a different process
HANDLE session_event;
//stores the TCB id
unsigned long tcb_id;
//to allow different threads to use this library simultaneously (thread-safe)
HANDLE hMutex;


/**
 * Waits for mutex
 */
void lock() {
	WaitForSingleObject(hMutex, INFINITE);
}


/**
 * Releases mutex
 */
void unlock() {
	ReleaseMutex(hMutex);
}




/**
 * Opens the TCB session. This method should be called
 * first of all. 
 * In this version, only 1 session is allowed per machine.
 * 
 * @return 0 if OK, -1 if TCB is stopped, 
 *					-2 if a session is already open,
 *					-3 if other error
 */
TCB_SERVICES_LIBRARY_API int open_tcb_session(void) {

	//create event for TCB to wait for our requests
	session_event = CreateEvent(NULL, TRUE, FALSE, L"session_event");
	if (session_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//if event is already created, a session is already being used
	if (GetLastError()==ERROR_ALREADY_EXISTS) {
		return -2;
	}

	//create event for TCB to wait for our requests
	tcb_wait_event = CreateEvent(NULL, TRUE, FALSE, L"tcb_wait_event");
	if (tcb_wait_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//create event for waiting for TCB answers
	dll_wait_event = CreateEvent(NULL, TRUE, FALSE, L"dll_wait_event");
	if (dll_wait_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//create event for waiting for TCB distributed 
	//timing failure detection service info
	wait_info_event = CreateEvent(NULL, TRUE, FALSE, L"wait_info_event");
	if (wait_info_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//if event isn't already created, TCB is not running
	if (GetLastError()!=ERROR_ALREADY_EXISTS) {
		return -1;
	}
	//unset event because it can be set (from a previous session)
	ResetEvent(wait_info_event);

	//create memory mapping TCB Services<->DLL
	fmap = CreateFileMapping(INVALID_HANDLE_VALUE, 
		NULL, PAGE_READWRITE, 0, sizeof(service_request_response_t), 
		L"tcb_memory_map");

	if (fmap==NULL) {
		//_wperror(L"Error creating memory mapping\n");
		//Sleep(5000);
		return -1;
	}
	
	//get pointer to memory map
	srrt = (service_request_response_t *) 
			MapViewOfFile(fmap, FILE_MAP_WRITE, 0, 0, 0);
	
	if (srrt==NULL) {
		//_wperror(L"Error mapping memory\n");
		//Sleep(5000);
		return -1;
	}
	//fill with zeros
	memset(srrt, 0, sizeof(service_request_response_t));


	//create a memory mapping to be used in DistFDService->TCB DLL waitInfo() comm
	HANDLE fmap1 = CreateFileMapping(INVALID_HANDLE_VALUE, 
		NULL, PAGE_READWRITE, 0, sizeof(fail_info_to_app_t), 
		L"wait_info_memory_map");
	if (fmap1==NULL) {
		//_wperror(L"Error creating memory mapping 2\n");
		return -1;
	}
	//associate the fail_info_to_app_t with the memory mapping
	fi =(fail_info_to_app_t *) 
			MapViewOfFile(fmap1, FILE_MAP_WRITE, 0, 0, 0);
	if (fi==NULL) {
		//_wperror("Error mapping memory 2\n");
		return -1;
	}
	//fill with zeros
	memset(fi, 0, sizeof(fail_info_to_app_t));

	//for thread-safety
	hMutex = CreateMutex(NULL, FALSE, NULL);
	if (hMutex==NULL) {
		return -3;
	}
	
	opened=TRUE;
	tcb_id = get_tcb_id();

	return 0;	
}


/**
 * Closes the TCB session.
 * 
 * @return 0 if OK, -1 if error
 */
TCB_SERVICES_LIBRARY_API int close_tcb_session(void){
	lock();
	//unmap memory
	if(!UnmapViewOfFile(srrt)){
		//_wperror(L"Error unmaping view of file\n");
		Sleep(5000);
		unlock();
		return -1;
	}
	//close memory map handle
	CloseHandle(fmap);
	//close event handles
	CloseHandle(dll_wait_event);
	CloseHandle(tcb_wait_event);
	opened=FALSE;
	unlock();
	return 0;
}


/**
 * Calls the tcb service: subscribe_event_channel
 *
 * Subscribe to a certain event channel.
 *
 * @param ec_id Id of the channel to subscribe (0<=ec_id<=31)
 *
 * @return 0 if OK,
 *		   -1 if already subscribed,
 *		   -2 if ec_id is invalid.
 */
TCB_SERVICES_LIBRARY_API int subscribe_tcb_event_channel(char ec_id) {
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}

	//call SUBSCRIBE_EVENT_CHANNEL
	srrt->request.service_id=SUBSCRIBE_EVENT_CHANNEL;
	srrt->request.ec_id=ec_id;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int ret = srrt->response.ret;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	unlock();
	//return answer
	return ret;
}



/**
 * Calls the tcb service: unsubscribe_event_channel
 *
 * Unsubscribe from a certain event channel.
 *
 * @param ec_id Id of the channel to unsubscribe (0<=ec_id<=31)
 *
 * @return 0 if OK,
 *		   -1 if not subscribed,
 *		   -2 if ec_id is invalid.
 */
TCB_SERVICES_LIBRARY_API int unsubscribe_tcb_event_channel(char ec_id) {
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}

	//call UNSUBSCRIBE_EVENT_CHANNEL
	srrt->request.service_id=UNSUBSCRIBE_EVENT_CHANNEL;
	srrt->request.ec_id=ec_id;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int ret = srrt->response.ret;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	unlock();
	//return answer
	return ret;
}



/**
 * Calls the tcb service: get_timestamp
 * 
 * Returns the local clock value (milliseconds).
 * 
 * @return Local clock value (milliseconds) if OK,
 *		   -1 if no TCB session is opened		   
 */
TCB_SERVICES_LIBRARY_API int get_tcb_timestamp(void){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}

	//call GET_TIMESTAMP
	srrt->request.service_id=GET_TIMESTAMP;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int timestamp = srrt->response.value;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	unlock();
	//return answer
	return timestamp;
}


/**
 * Calls the tcb service: get_global_timestamp
 * 
 * Returns the global clock value (milliseconds) at local instant specified.
 * If localInstant is 0, the present global clock value is returned.
 * 
 * @param localInstant A timestamp obtained with get_timestamp() or 0.
 * 
 * @return Global clock value (milliseconds) at local instant specified if OK
 * 		   (If localInstant is 0, the present global clock value is returned),
 *			-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int get_tcb_global_timestamp(int localInstant){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}
	//call GET_GLOBAL_TIMESTAMP
	srrt->request.service_id=GET_GLOBAL_TIMESTAMP;
	srrt->request.localInstant = localInstant;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int timestamp = srrt->response.value;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	unlock();
	//return answer
	return timestamp;
}


/**
 * Calls the tcb pseudo-service: get_tcb_id
 * 
 * Returns the id of the TCB we are connected to.
 * 
 * @return Id of the TCB we are connected to if OK,
 *		   -1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API unsigned long get_tcb_id(void){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}
	//call GET_TCB_ID
	srrt->request.service_id=GET_TCB_ID;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	unsigned long id = srrt->response.id;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	unlock();
	//return answer
	return id;
}


/**
 * Calls the tcb service: start_local_measurement
 * 
 * Starts a local measurement at instant start_ev.
 * 
 * @param start_ev Instant that marks the begin of the measurement
 * 
 * @return Tag that identifies the measurement if OK,
 *		   -1 if no tag is available,
 *		   -2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_measurement(int start_ev){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -2;
	}
	//call START_MEASUREMENT
	srrt->request.service_id=START_MEASUREMENT;
	srrt->request.start_ev = start_ev;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int tag = srrt->response.tag;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	unlock();
	//return answer
	return tag;	
}


/**
 * Calls the tcb service: end_local_measurement
 * 
 * Stops the measurement identified by tag.
 * 
 * @param tag Measurement identifier
 *
 * @return Information about measurement if OK,
 *		   tcb_measurement_info_t.end_ev=-1 if tag is invalid,
 *		   tcb_measurement_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_measurement_info_t end_tcb_local_measurement(int tag){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		tcb_measurement_info_t ret;
		ret.end_ev = -2;
		unlock();
		return ret;
	}
	//call END_MEASUREMENT
	srrt->request.service_id=END_MEASUREMENT;
	srrt->request.tag = tag;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	
	//construct return value
	tcb_measurement_info_t ret;
	ret.end_ev = srrt->response.mi.end_ev;
	ret.procdelay = srrt->response.mi.procdelay;
	unlock();
	return ret;
}


/**
 * Calls the tcb service: timely_exec
 * 
 * Starts the execution of an application function. 
 * TCB verifies if it's possible to timely execute it.
 * 
 * @param start_ev A time instant
 * @param delay start_ev+delay marks the exec begin
 * @param max_exec Deadline for the execution, counting since start_ev
 * @param dllName Name of the DLL file where the function is defined
 * @param funcName Function name 
 * @param wcetName Worst Case Execution Time function name
 * 
 * @return The time instant when the function ends if OK;
 *			-1 if no available entry in the executions table
 *			-2 if dll not found
 *			-3 if function not found
 *			-4 if wcet function not found
 *			-5 if it's not possible to execute function
 *			-6 if it's not possible to timely execute function
 *			-7 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int tcb_timely_exec(int start_ev, int delay, int max_exec, char* dllName,
										 char* funcName, char* wcetName){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -7;
	}
	//call TIMELY_EXEC
	srrt->request.service_id=TIMELY_EXEC;
	srrt->request.execArguments.start_ev=start_ev;
	srrt->request.execArguments.delay = delay;
	srrt->request.execArguments.max_exec = max_exec;
	//convert dllName to wchar_t
	wsprintf(srrt->request.execArguments.dllName, L"%S", dllName);
	//convert funcName to wchar_t
	wsprintf(srrt->request.execArguments.funcName, L"%S", funcName);
	//convert wcetName to wchar_t
	wsprintf(srrt->request.execArguments.wcetName, L"%S", wcetName);
	
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int end_ev = srrt->response.end_ev;
	srrt->type=0;
	unlock();
	//return answer
	return end_ev;
}


/**
 * Calls the tcb service: start_local_fd
 *
 * Starts a local timing failure detection.
 * 
 * @param start_ev A time instant
 * @param spec A timing failure occurs if start_ev+spec instant is reached
 *	            before a call endLocalFD
 * @param deadline Deadline for the execution of handler, if a failure occurs
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 *
 * @return Tag that identifies the timing failure detection if OK,
 *			-1 if no available entry in the localFD table,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if it's not possible to execute timing failure detection,
 *			-6 if a timing failure already occured,
 *			-7 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_fd(int start_ev, int spec, int deadline, 
											char* dllName, char* funcName, char* wcetName){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -7;
	}
	//call START_LOCAL_FD
	srrt->request.service_id=START_LOCAL_FD;
	srrt->request.localFDArguments.start_ev=start_ev;
	srrt->request.localFDArguments.spec = spec;
	srrt->request.localFDArguments.deadline = deadline;
	//convert dllName to wchar_t
	wsprintf(srrt->request.localFDArguments.dllName, L"%S", dllName);
	//convert funcName to wchar_t
	wsprintf(srrt->request.localFDArguments.funcName, L"%S", funcName);
	//convert wcetName to wchar_t
	wsprintf(srrt->request.localFDArguments.wcetName, L"%S", wcetName);
    
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int tag = srrt->response.tag;
	srrt->type=0;
	unlock();
	//return answer
	return tag;
}


/**
 * Calls the tcb service: end_local_fd
 *
 * Ends a local timing failure detection.
 * 
 * @param tag Local timing detection identifier
 *
 * @return Information about the timing failure detection if OK,
 *			local_fd_info_t.end_ev=-1 if tag is invalid,
 *			local_fd_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_local_fd_info_t end_tcb_local_fd(int tag){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		tcb_local_fd_info_t ret;
		ret.end_ev=-2;
		unlock();
		return ret;
	}
	//call END_LOCAL_FD
	srrt->request.service_id=END_LOCAL_FD;
	srrt->request.tag = tag;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);

	//construct return value
	tcb_local_fd_info_t ret;
	ret.end_ev = srrt->response.endLocalFDRet.end_ev;
	ret.faulty = srrt->response.endLocalFDRet.faulty;
	ret.procdelay = srrt->response.endLocalFDRet.procdelay;
	unlock();
	return ret;
}


/**
 * Waits for info from TCB distributed timing failure detection service.
 *
 * @return Information about the distributed timing failure detection if OK,
 *		   tcb_distributed_fd_info_t.from=-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_fd_info_t tcb_wait_info(void){
	
	if (!opened) {
		printf("Open a TCB session first!\n");
		tcb_distributed_fd_info_t ret;
		ret.from = -1;
		return ret;
	}
	//wait for the info
	WaitForSingleObject(wait_info_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(wait_info_event);

	//construct DistributedFDInfo to return
	tcb_distributed_fd_info_t dfi;
	dfi.from = fi->from;
	dfi.seq = fi->seq;
	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (!fi->failure_info[i].empty) {
			dfi.failInfo[i].empty = FALSE;
			dfi.failInfo[i].id = fi->failure_info[i].id;
			dfi.failInfo[i].fail = fi->failure_info[i].fail;
			dfi.failInfo[i].latenessAvail = fi->failure_info[i].lateness_avail;
			dfi.failInfo[i].latenessDegree = fi->failure_info[i].lateness_degree;
		}
		else 
			dfi.failInfo[i].empty = TRUE;
	}
	return dfi;
}


/**
 * Calls the TCB pseudo-service: define_remote_fd_handler
 *
 * GOAL: to define the handler to be used on remote failure detection.
 *
 * PARAM:
 *		dllName Name of the DLL file where the handler is defined
 *		funcName Handler name 
 *		wcetName Handler's Worst Case Execution Time function name
 *		deadline Maximum exec time of failure handler (if a failure occurs)
 *
 * RETURNS: 
 *		0 if OK,
 *		-2 if dll not found,
 *		-3 if function not found,
 *		-4 if wcet function not found,
 *		-5 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int define_tcb_remote_fd_handler(char* dllName, char* funcName, 
														  char* wcetName, int deadline){
	lock();
	if (!opened) {
		printf("Open a TCB session first!\n");
		unlock();
		return -5;
	}
	//call DEFINE_REMOTE_HANDLER
	srrt->request.service_id=DEFINE_REMOTE_HANDLER;
	//convert dllName to wchar_t
	wsprintf(srrt->request.remote_handler.dllName, L"%S", dllName);    
	//convert funcName to wchar_t	
	wsprintf(srrt->request.remote_handler.funcName, L"%S", funcName);    
	//convert wcetName to wchar_t	
	wsprintf(srrt->request.remote_handler.wcetName, L"%S", wcetName);
    
	//set deadline
	srrt->request.remote_handler.deadline = deadline;
	
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int ret = srrt->response.ret;
	srrt->type=0;
	unlock();
	//return answer
	return ret;		
}


/**
 * Calls the tcb service: start_dist_measurement
 * 
 * Starts a distributed measurement at instant start_ev.
 * 
 * @param IN start_ev Instant that marks the begin of the measurement.
 * @param OUT action_id Id that identifies the distributed action.
 * 
 * @return  0 if OK,
 *		   -1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API 
int start_tcb_dist_measurement(int start_ev, tcb_dist_action_id_t *action_id) {
	lock();
	if (!opened){
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}
	unlock();
	return fill_action_id(start_ev, 0, action_id);
}


/**
 * Calls the TCB service: start_dist_fd
 * 
 * Starts simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the sender.
 *
 * If detection is not ended in all recipients subscribed to event_channel_id 
 * until start_ev+spec,
 * the specified failure handler is executed (in the sender) with the maximum
 * exec time specified.
 *
 * @param IN start_ev the initial time instant of the failure detection and measurement
 * @param IN event_channel_id Event channel where the timing failure detection will be
 *						   started (0<=event_channel_id<=31)
 * @param IN deadline Maximum exec time of failure handler (if a failure occurs)
 * @param IN dllName Name of the DLL file where the failure handler is defined
 * @param IN funcName Failure handler name
 * @param IN wcetName Failure handler Worst Case Execution Time function name
 * @param OUT action_id Id that identifies the distributed action.
 *
 * @return 0 if OK, 
 *		   -1 if no TCB session is opened,
 *		   -2 if dll not found,
 *		   -3 if function not found,
 *		   -4 if wcet function not found,
 *		   -5 if there are too many distributed fds already running,
 *		   -6 if a timing failure already occured,
 *		   -7 if it's not possible to execute timing failure detection,
 *		   -8 if no TCB is subscribed to event channel specified. 
 */
TCB_SERVICES_LIBRARY_API 
int start_tcb_dist_fd(int start_ev, char event_channel_id, int spec, int deadline,
					  char* dllName, char* funcName, char* wcetName, 
					  tcb_dist_action_id_t *action_id) {
	lock();
	if (!opened){
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}

	int seq=0;
	//if there is at least one destination, call start_distributed_fd
	//if (dest_mask!=0){
	seq = start_tcb_distributed_fd(event_channel_id, start_ev, spec, deadline, 
										dllName, funcName, wcetName);
	if (seq<0) {
		unlock();
		return seq;
	}

	//fill with TCB id
	action_id->tcb_id= tcb_id;
	//}
	unlock();
	//fill with other values
	return fill_action_id(start_ev, seq, action_id);
}


/**
 * Calls the TCB service: start_dist_remote_fd
 * 
 * Starts simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the recipients.
 *
 * For each recipient subscribed to event_channel_id: if service is not ended 
 * until start_ev+spec,
 * the failure handler specified in define_tcb_remote_fd_handler
 * is executed with the maximum exec time specified in that function.
 *
 * @param IN start_ev the initial time instant of the failure detection and measurement
 * @param IN event_channel_id Event channel where the timing failure detection will be
 *						   started (0<=event_channel_id<=31)
 * @param IN spec Maximum duration starting at start_ev (for failure detection)
 * @param OUT action_id Id that identifies the distributed action.
 *
 * @return 0 if OK, 
 *		   -1 if no TCB session is opened,
 *		   -5 if there are too many distributed fds already running,
 *		   -6 if a timing failure already occured,
 *		   -7 if it's not possible to execute timing failure detection
 */
TCB_SERVICES_LIBRARY_API 
int start_tcb_dist_remote_fd(int start_ev, char event_channel_id, int spec, 
							 tcb_dist_action_id_t *action_id) {
	lock();
	if (!opened){
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}

	int seq=0;
	//if there is at least one destination, call start_distributed_remote_fd
	//if (dest_mask!=0){
	seq = start_tcb_distributed_remote_fd(event_channel_id, start_ev, spec);
	//if an error occured, return now
	if (seq<0) {
		unlock();
		return seq;
	}
	//fill action_id with TCB id
	action_id->tcb_id = tcb_id;
	//}
	unlock();
	//fill remainder of action_id
	return fill_action_id(start_ev, seq, action_id);
}


/**
 * Calls the tcb service: end_dist_action
 * 
 * Ends the distributed action (measurement/fd/remote fd)
 * identified by action_id.
 * 
 * @param IN action_id Distributed action id.
 * @param OUT measurement_info Info about the distributed measurement result.
 *
 * @return 0 if OK,
 *		   -1 if no TCB session is opened,
 *		   -2 if cannot end distributed FD.
 */
TCB_SERVICES_LIBRARY_API 
int end_tcb_dist_action(tcb_dist_action_id_t *action_id, 
						tcb_dist_measurement_info_t *measurement_info) {
	lock();			
	if (!opened){
		printf("Open a TCB session first!\n");
		unlock();
		return -1;
	}

	//get a timestamp from TCB
	measurement_info->end_ev = get_tcb_timestamp();
	//get the global timestamp to calculate the msg delay
	int end_ev_global = get_tcb_global_timestamp(measurement_info->end_ev);
	//calculate the delay
	measurement_info->delay = end_ev_global - action_id->start_ev_global;
	measurement_info->error=DISTRIBUTED_MEASUREMENT_ERROR;
	if (action_id->seq!=0) {
		boolean ok;
		//if msg was sent with FD, notify TCB
		ok=tcb_msg_rcv_on_payload(action_id->seq, action_id->tcb_id, 
								  measurement_info->end_ev);
		if (!ok) {
			unlock();
			return -2;
		}
	}
	unlock();
	return 0;
}
	



/*******************
 * PRIVATE FUNCTIONS
 *******************/

/**
 * Private function that fills action_id.
 * Called by the various public start_dist().
 *
 * @param IN start_ev the initial time instant of the action to be measured.
 * @param IN seq sequence number to put in action_id.
 * @param OUT action_id Id that identifies the distributed action.
 *
 * @return 0 if OK
 */
int fill_action_id (int start_ev, int seq, tcb_dist_action_id_t *action_id) {
	//get global timestamp corresponding to local start_ev
	int start_ev_global = get_tcb_global_timestamp(start_ev);
	
	//set start_ev
	action_id->start_ev_global = start_ev_global;
	//set seq
	action_id->seq = seq;
	
	return 0;
}





/**
 * Calls the tcb service: start_distributed_fd
 * 
 * Starts a distributed timing failure detection.
 * 
 * @param event_channel_id Event channel where the timing failure detection will be
 *						   started (0<=event_channel_id<=31)
 * @param sendEv The initial time instant of the failure detection.
 * @param spec Maximum duration starting at sendEv.
 * @param deadline Maximum exec time of failure handler (if a failure occurs)
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 *
 * @return seq that identifies the dist. timing failure detection,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection,
 *			-8 if no TCB is subscribed to event channel specified. 
 */ 
int start_tcb_distributed_fd(char event_channel_id, int sendEv, int spec, 
						 int deadline, char* dllName, char* funcName, 
						 char* wcetName) {
	//call START_DISTRIBUTED_FD
	srrt->request.service_id=START_DISTRIBUTED_FD;
	srrt->request.distFDArguments.ec_id = event_channel_id;
	srrt->request.distFDArguments.start_ev = sendEv;
	srrt->request.distFDArguments.spec = spec;
	srrt->request.distFDArguments.deadline = deadline;
	//convert dllName to wchar_t
	wsprintf(srrt->request.distFDArguments.dllName, L"%S", dllName);
	//convert funcName to wchar_t
	wsprintf(srrt->request.distFDArguments.funcName, L"%S", funcName);    
	//convert wcetName to wchar_t
	wsprintf(srrt->request.distFDArguments.wcetName, L"%S", wcetName);
    	
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int seq = srrt->response.seq;
	srrt->type=0;
	//return answer
	return seq;		
}


/**
 * Calls the tcb service: start_distributed_remote_fd
 * 
 * Starts a distributed remote timing failure detection.
 * 
 * @param event_channel_id Event channel where the timing failure detection will be
 *						   started (0<=event_channel_id<=31)
 * @param sendEv The initial time instant of the failure detection.
 * @param spec Maximum duration starting at sendEv.
 *
 * @return seq that identifies the dist. timing failure detection,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection. 
 */ 
int start_tcb_distributed_remote_fd(char event_channel_id, int sendEv, int spec) {
	//call START_DISTRIBUTED_REMOTE_FD
	srrt->request.service_id=START_DISTRIBUTED_REMOTE_FD;
	srrt->request.distFDArguments.ec_id = event_channel_id;
	srrt->request.distFDArguments.start_ev = sendEv;
	srrt->request.distFDArguments.spec = spec;
		
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int seq = srrt->response.seq;
	srrt->type=0;
	//return answer
	return seq;		
}


/**
 * PSEUDO-SERVICE: msg_rcv_on_payload
 *
 * GOAL: to DLL notify TCB that a msg was received on payload
 * 
 * PARAM: 
 *		seq - sequence number
 *		from - sender id
 *		time_of_receipt - reception time instant
 *
 * RETURNS : TRUE if msg can be delivered,
 *					  FALSE if msg cannot be delivered			   
 */
bool tcb_msg_rcv_on_payload(int seq, unsigned long from, int time_of_receipt){
	//call MSG_RCV_ON_PAYLOAD
	srrt->request.service_id=MSG_RCV_ON_PAYLOAD;
	srrt->request.msgRcvOnPayloadArguments.seq=seq;
	srrt->request.msgRcvOnPayloadArguments.from=from;
	srrt->request.msgRcvOnPayloadArguments.time_of_receipt=time_of_receipt;
		
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	bool deliver = srrt->response.deliver;
	srrt->type=0;
	//return answer
	return deliver;		
}


/*******************************
 * Auxiliary Functions
 *******************************/

/**
 * Desserializes a char array containing an integer.
 * 
 * @param b the char array to desserialize
 * @param off start offset
 * @return the desserialized integer.
 */
int byteArrayToInt(char* b, int off) {
	int ret = 0;
	char* ret_pt = (char*)&ret;

	ret_pt[0] = b[off];
	ret_pt[1] = b[off + 1]; 
	ret_pt[2] = b[off + 2]; 
	ret_pt[3] = b[off + 3];

	return ret;
}


/**
 * Serializes an integer into a char array
 * 
 * @param i the int to serialize
 * @param b to return the serialization
 * @param off start offset
 */
void intToByteArray(int* i, char* b, int off) {
	
	char* int_pt = (char*)i;
	

	b[off+0] = int_pt[0];
	b[off+1] = int_pt[1];
	b[off+2] = int_pt[2];
	b[off+3] = int_pt[3];
}

